@@ -1,7 +1,7 @@
 //      fm-dir-tree-model.c
 //
 //      Copyright 2010 Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
-//      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+//      Copyright 2012-2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
 //
 //      This program is free software; you can redistribute it and/or modify
 //      it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  * @short_description: A model for directory tree view
  * @title: FmDirTreeModel
  *
- * @include: libfm/fm-dir-tree-model.h
+ * @include: libfm/fm-gtk.h
  *
  * The #FmDirTreeModel represents tree of folders which can be used by
  * #FmDirTreeView to create tree-like expandable list of directories.
@@ -33,6 +33,8 @@
 #include <config.h>
 #endif
 
+#define FM_DISABLE_SEAL
+
 #include "fm-dir-tree-model.h"
 #include "fm-folder.h"
 #include "fm-icon-pixbuf.h"
@@ -415,7 +417,7 @@ static void fm_dir_tree_model_get_value ( GtkTreeModel *tree_model,
             /* parent is always non NULL. otherwise it's a bug. */
             FmDirTreeItem* parent = (FmDirTreeItem*)item->parent->data;
             if(parent->folder && fm_folder_is_loaded(parent->folder))
-                g_value_set_string( value, _("<No Sub Folder>"));
+                g_value_set_string( value, _("<No subfolders>"));
             else
                 g_value_set_string( value, _("Loading..."));
         }
@@ -488,7 +490,6 @@ static gboolean fm_dir_tree_model_iter_has_child(GtkTreeModel *tree_model,
 {
     GList* item_l;
     FmDirTreeItem* item;
-    /* FIXME: is NULL iter allowed here? */
     g_return_val_if_fail( iter != NULL, FALSE );
     g_return_val_if_fail( iter->stamp == FM_DIR_TREE_MODEL(tree_model)->stamp, FALSE );
 
@@ -680,10 +681,8 @@ static GList* insert_file_info(FmDirTreeModel* model, GList* parent_l, GtkTreePa
     FmDirTreeItem* parent_item = (FmDirTreeItem*)parent_l->data;
     FmDirTreeItem* item = fm_dir_tree_item_new(model, parent_l);
     item->fi = fm_file_info_ref(fi);
-    FmPath* path = fm_file_info_get_path(fi);
 
-    /* fm_file_info_is_hidden() is slower here*/
-    if(!model->show_hidden && fm_path_get_basename(path)[0] == '.') /* hidden folder */
+    if(!model->show_hidden && fm_file_info_is_hidden(fi)) /* hidden folder */
     {
         parent_item->hidden_children = g_list_prepend(parent_item->hidden_children, item);
         item_l = parent_item->hidden_children;
@@ -693,7 +692,8 @@ static GList* insert_file_info(FmDirTreeModel* model, GList* parent_l, GtkTreePa
     return item_l;
 }
 
-static void remove_item(FmDirTreeModel* model, GList* item_l)
+/* deletes item from lists but not frees data */
+static void remove_item_l(FmDirTreeModel* model, GList* item_l)
 {
     GtkTreePath* tp;
     FmDirTreeItem* item = (FmDirTreeItem*)item_l->data;
@@ -701,44 +701,49 @@ static void remove_item(FmDirTreeModel* model, GList* item_l)
     g_return_if_fail(item != NULL);
     tp = item_to_tree_path(model, item_l);
 
-        if(item->parent)
-        {
-            FmDirTreeItem* parent_item = item->parent->data;
+    if(item->parent)
+    {
+        FmDirTreeItem* parent_item = item->parent->data;
 
-            parent_item->children = g_list_delete_link(parent_item->children, item_l);
-            /* signal the view that we removed the item. */
-            gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), tp);
-            /* If the item being removed is the last child item of parent_item,
-             * we need to insert a place holder item to keep it expandable. */
-            if(parent_item->children == NULL)
+        parent_item->children = g_list_delete_link(parent_item->children, item_l);
+        /* signal the view that we removed the item. */
+        gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), tp);
+        /* If the item being removed is the last child item of parent_item,
+         * we need to insert a place holder item to keep it expandable. */
+        if(parent_item->children == NULL)
+        {
+            GList* parent_l = item->parent;
+            gtk_tree_path_up(tp);
+            if(fm_config->no_child_non_expandable)
             {
-                GList* parent_l = item->parent;
-                //int idx = gtk_tree_path_get_indices(tp)[gtk_tree_path_get_depth(tp) - 1];
-                gtk_tree_path_up(tp);
-                if(fm_config->no_child_non_expandable)
-                {
-                    GtkTreeIter it;
-                    item_to_tree_iter(model, parent_l, &it);
-                    /* signal the view to redraw row removing expander */
-                    gtk_tree_model_row_has_child_toggled(GTK_TREE_MODEL(model), tp, &it);
-                }
-                else
-                    add_place_holder_child_item(model, parent_l, tp, TRUE);
-                //gtk_tree_path_append_index(tp, 0);
+                GtkTreeIter it;
+                item_to_tree_iter(model, parent_l, &it);
+                /* signal the view to redraw row removing expander */
+                gtk_tree_model_row_has_child_toggled(GTK_TREE_MODEL(model), tp, &it);
             }
+            else
+                add_place_holder_child_item(model, parent_l, tp, TRUE);
         }
-        else /* root item */
-        {
-            /* FIXME: this needs more testing. */
-            model->roots = g_list_delete_link(model->roots, item_l);
-            /* signal the view that we removed the item. */
-            gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), tp);
-        }
+    }
+    else /* root item */
+    {
+        /* FIXME: this needs more testing. */
+        model->roots = g_list_delete_link(model->roots, item_l);
+        /* signal the view that we removed the item. */
+        gtk_tree_model_row_deleted(GTK_TREE_MODEL(model), tp);
+    }
 
-    fm_dir_tree_item_free(item, item_l); /* item_l is freed already */
     gtk_tree_path_free(tp);
 }
 
+/* deletes and frees item with data */
+static void remove_item(FmDirTreeModel* model, GList* item_l)
+{
+    FmDirTreeItem* item = (FmDirTreeItem*)item_l->data;
+    remove_item_l(model, item_l);
+    fm_dir_tree_item_free(item, item_l); /* item_l is freed already */
+}
+
 /* find child item by filename, and retrive its index if idx is not NULL. */
 static GList* children_by_name(FmDirTreeModel* model, GList* children, const char* name, int* idx)
 {
@@ -873,19 +878,10 @@ static void on_folder_files_added(FmFolder* folder, GSList* files, GList* item_l
     for(l = files; l; l = l->next)
     {
         FmFileInfo* fi = FM_FILE_INFO(l->data);
-        /* FIXME: should FmFolder generate "files-added" signal on
-         * its first-time loading? Isn't "loaded" signal enough? */
-        if(fm_file_info_is_dir(fi)) /* TODO: maybe adding files can be allowed later */
+        if(fm_file_info_is_dir(fi)) /* FIXME: maybe adding files can be allowed later */
         {
-            /* ensure that the file is not yet in our model
-            FmPath* path = fm_file_info_get_path(fi); */
-            /* 
-            * Ideally FmFolder should not emit files-added signals for files that
-            * already exists. So there is no need to check for duplication here.
-            GList* new_item_l = children_by_name(model, item->children, path->name, NULL);
-            if(!new_item_l)
-
-            GList* new_item_l = */
+            /* Ideally FmFolder should not emit files-added signals for files that
+             * already exists. So there is no need to check for duplication here. */
             insert_file_info(model, item_l, tp, fi);
         }
     }
@@ -914,7 +910,7 @@ static void on_folder_files_changed(FmFolder* folder, GSList* files, GList* item
     GSList* l;
     FmDirTreeItem* item = (FmDirTreeItem*)item_l->data;
     FmDirTreeModel* model = item->model;
-    /* GtkTreePath* tp = item_to_tree_path(model, item_l); */
+    GtkTreePath* tp = item_to_tree_path(model, item_l);
 
     /* g_debug("files changed!!"); */
 
@@ -929,16 +925,21 @@ static void on_folder_files_changed(FmFolder* folder, GSList* files, GList* item
         if(changed_item_l)
         {
             FmDirTreeItem* changed_item = (FmDirTreeItem*)changed_item_l->data;
+            GtkTreeIter it;
             if(changed_item->fi)
                 fm_file_info_unref(changed_item->fi);
             changed_item->fi = fm_file_info_ref(fi);
-            /* FIXME: inform gtk tree view about the change */
+            /* inform gtk tree view about the change */
+            item_to_tree_iter(model, changed_item_l, &it);
+            gtk_tree_path_append_index(tp, idx);
+            gtk_tree_model_row_changed(GTK_TREE_MODEL(model), tp, &it);
+            gtk_tree_path_up(tp);
 
             /* FIXME and TODO: check if we have sub folder */
             /* item_queue_subdir_check(model, changed_item_l); */
         }
     }
-    /* gtk_tree_path_free(tp); */
+    gtk_tree_path_free(tp);
 }
 
 static inline void item_free_folder(FmFolder* folder, gpointer item_l)
@@ -1207,7 +1208,7 @@ const char* fm_dir_tree_row_get_disp_name(FmDirTreeModel* model, GtkTreeIter* it
     /* parent is always non NULL. otherwise it's a bug. */
     parent = (FmDirTreeItem*)item->parent->data;
     if(parent->folder && fm_folder_is_loaded(parent->folder))
-        return _("<No Sub Folder>");
+        return _("<No subfolders>");
     return _("Loading...");
 }
 
@@ -1237,41 +1238,85 @@ gboolean fm_dir_tree_row_is_loaded(FmDirTreeModel* model, GtkTreeIter* iter)
     return item->loaded;
 }
 
-#if 0
-static void item_show_hidden_children(FmDirTreeModel* model, GList* item_l, gboolean show_hidden)
+static void item_hide_hidden_children(FmDirTreeModel *model, GList *item_l)
 {
-    FmDirTreeItem* item = (FmDirTreeItem*)item_l->data;
-    GList* child_l;
-    /* TODO: show hidden items */
-    if(show_hidden)
-    {
-        while(item->hidden_children)
-        {
+    FmDirTreeItem *item = (FmDirTreeItem*)item_l->data;
+    FmDirTreeItem *child;
+    GList *child_l, *next;
 
-        }
-    }
-    else
+    if (item) for (child_l = item->children; child_l; child_l = next)
     {
-        while(item->children)
+        next = child_l->next;
+        child = child_l->data;
+        if (G_UNLIKELY(child->fi == NULL)) /* placeholder */
+            continue;
+        if (fm_file_info_is_hidden(child->fi))
         {
-
+            /* remove from visibility in model */
+            remove_item_l(model, child_l);
+            /* do cleanup on item data */
+            if (child->folder)
+                item_free_folder(child->folder, child_l);
+            child->folder = NULL;
+            child->expanded = FALSE;
+            child->loaded = FALSE;
+            if(child->children)
+            {
+                _g_list_foreach_l(child->children, (GFunc)fm_dir_tree_item_free_l, NULL);
+                g_list_free(child->children);
+                child->children = NULL;
+            }
+            if(child->hidden_children)
+            {
+                _g_list_foreach_l(child->hidden_children, (GFunc)fm_dir_tree_item_free_l, NULL);
+                g_list_free(child->hidden_children);
+                child->hidden_children = NULL;
+            }
+            /* item is clean so can be added to hidden children */
+            item->hidden_children = g_list_prepend(item->hidden_children, child);
         }
+        else
+            item_hide_hidden_children(model, child_l); /* do recursion */
     }
 }
 
+static void item_show_hidden_children(FmDirTreeModel *model, GList *item_l)
+{
+    FmDirTreeItem *item = (FmDirTreeItem*)item_l->data;
+    FmDirTreeItem *child;
+    GtkTreePath *tp;
+    GList *child_l;
+
+    tp = item_to_tree_path(model, item_l);
+    for (child_l = item->children; child_l; child_l = child_l->next)
+        item_show_hidden_children(model, child_l); /* do recursion */
+    while (item->hidden_children)
+    {
+        /* isolate child */
+        child = item->hidden_children->data;
+        item->hidden_children = g_list_delete_link(item->hidden_children,
+                                                   item->hidden_children);
+        /* insert it into visible list */
+        insert_item(model, item_l, tp, child);
+    }
+    gtk_tree_path_free(tp);
+}
+
 void fm_dir_tree_model_set_show_hidden(FmDirTreeModel* model, gboolean show_hidden)
 {
+    GList *l;
+
     if(show_hidden != model->show_hidden)
     {
         /* filter the model to hide hidden folders */
         if(model->show_hidden)
-        {
-
-        }
+            for (l = model->roots; l; l = l->next)
+                item_hide_hidden_children(model, l);
+        /* filter the model to show hidden folders */
         else
-        {
-
-        }
+            for (l = model->roots; l; l = l->next)
+                item_show_hidden_children(model, l);
+        model->show_hidden = show_hidden;
     }
 }
 
@@ -1280,6 +1325,7 @@ gboolean fm_dir_tree_model_get_show_hidden(FmDirTreeModel* model)
     return model->show_hidden;
 }
 
+#if 0
 /* TODO: */ void fm_dir_tree_model_reload(FmDirTreeModel* model)
 {
 }
