@@ -3,6 +3,7 @@
  *
  *      Copyright 2009 PCMan <pcman@debian>
  *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2012 Vadim Ushakov <igeekless@gmail.com>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -25,7 +26,7 @@
  * @short_description: Different widgets and utilities that use GTK+
  * @title: Libfm-gtk utils
  *
- * @include: libfm/fm-gtk-utils.h
+ * @include: libfm/fm-gtk.h
  *
  */
 
@@ -47,7 +48,7 @@
 #include "fm-config.h"
 
 static GtkDialog*   _fm_get_user_input_dialog   (GtkWindow* parent, const char* title, const char* msg);
-static gchar*       _fm_user_input_dialog_run   (GtkDialog* dlg, GtkEntry *entry);
+static gchar*       _fm_user_input_dialog_run   (GtkDialog* dlg, GtkEntry *entry, GtkWidget *extra);
 
 /**
  * fm_show_error
@@ -254,7 +255,41 @@ gchar* fm_get_user_input(GtkWindow* parent, const char* title, const char* msg,
     if(default_text && default_text[0])
         gtk_entry_set_text(entry, default_text);
 
-    return _fm_user_input_dialog_run(dlg, entry);
+    return _fm_user_input_dialog_run(dlg, entry, NULL);
+}
+
+/**
+ * fm_get_user_input_n
+ * @parent: a window to place dialog over it
+ * @title: title for dialog window
+ * @msg: the message to present to the user
+ * @default_text: the default answer
+ * @n: which part of default answer should be selected
+ * @extra: additional widgets to display (can be focusable)
+ *
+ * Presents the message to user and retrieves entered text.
+ * In presented dialog the part of @default_text with length @n will be
+ * preselected for edition (n < 0 means select all).
+ * Returned data should be freed with g_free() after usage.
+ *
+ * Returns: (transfer full): entered text.
+ *
+ * Since: 1.2.0
+ */
+gchar* fm_get_user_input_n(GtkWindow* parent, const char* title, const char* msg,
+                           const char* default_text, gint n, GtkWidget* extra)
+{
+    GtkDialog* dlg = _fm_get_user_input_dialog( parent, title, msg);
+    GtkEntry* entry = GTK_ENTRY(gtk_entry_new());
+    gtk_entry_set_activates_default(entry, TRUE);
+
+    if(default_text && default_text[0])
+    {
+        gtk_entry_set_text(entry, default_text);
+        gtk_editable_select_region(GTK_EDITABLE(entry), 0, n);
+    }
+
+    return _fm_user_input_dialog_run(dlg, entry, extra);
 }
 
 /**
@@ -270,6 +305,8 @@ gchar* fm_get_user_input(GtkWindow* parent, const char* title, const char* msg,
  * Returns: (transfer full): entered text.
  *
  * Since: 0.1.0
+ *
+ * Deprecated: 1.2.0:
  */
 FmPath* fm_get_user_input_path(GtkWindow* parent, const char* title, const char* msg, FmPath* default_path)
 {
@@ -287,7 +324,7 @@ FmPath* fm_get_user_input_path(GtkWindow* parent, const char* title, const char*
         gtk_entry_set_text(entry, path_str);
     }
 
-    str = _fm_user_input_dialog_run(dlg, entry);
+    str = _fm_user_input_dialog_run(dlg, entry, NULL);
     path = fm_path_new_for_str(str);
 
     g_free(path_str);
@@ -335,7 +372,7 @@ static gchar* fm_get_user_input_rename(GtkWindow* parent, const char* title, con
         }
     }
 
-    return _fm_user_input_dialog_run(dlg, entry);
+    return _fm_user_input_dialog_run(dlg, entry, NULL);
 }
 
 static GtkDialog* _fm_get_user_input_dialog(GtkWindow* parent, const char* title, const char* msg)
@@ -349,13 +386,14 @@ static GtkDialog* _fm_get_user_input_dialog(GtkWindow* parent, const char* title
                                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                 GTK_STOCK_OK, GTK_RESPONSE_OK, NULL));
     GtkWidget* label = gtk_label_new(msg);
+    GtkBox* box = (GtkBox*)gtk_dialog_get_content_area(dlg);
     gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
 
     gtk_dialog_set_alternative_button_order(dlg, GTK_RESPONSE_OK, GTK_RESPONSE_CANCEL, -1);
-    gtk_box_set_spacing((GtkBox*)gtk_dialog_get_content_area(dlg), 6);
-    gtk_box_pack_start((GtkBox*)gtk_dialog_get_content_area(dlg), label, FALSE, TRUE, 6);
+    gtk_box_set_spacing(box, 6);
+    gtk_box_pack_start(box, label, FALSE, TRUE, 6);
 
-    gtk_container_set_border_width(GTK_CONTAINER((GtkBox*)gtk_dialog_get_content_area(dlg)), 12);
+    gtk_container_set_border_width(GTK_CONTAINER(box), 12);
     gtk_container_set_border_width(GTK_CONTAINER(dlg), 5);
     gtk_dialog_set_default_response(dlg, GTK_RESPONSE_OK);
     gtk_window_set_default_size(GTK_WINDOW(dlg), 480, -1);
@@ -363,17 +401,20 @@ static GtkDialog* _fm_get_user_input_dialog(GtkWindow* parent, const char* title
     return dlg;
 }
 
-static gchar* _fm_user_input_dialog_run(GtkDialog* dlg, GtkEntry *entry)
+static gchar* _fm_user_input_dialog_run(GtkDialog* dlg, GtkEntry *entry,
+                                        GtkWidget *extra)
 {
     char* str = NULL;
+    GtkBox *box = GTK_BOX(gtk_dialog_get_content_area(dlg));
     int sel_start, sel_end;
     gboolean has_sel;
 
     /* FIXME: this workaround is used to overcome bug of gtk+.
      * gtk+ seems to ignore select region and select all text for entry in dialog. */
     has_sel = gtk_editable_get_selection_bounds(GTK_EDITABLE(entry), &sel_start, &sel_end);
-    gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(dlg)),
-                       GTK_WIDGET(entry), FALSE, TRUE, 6);
+    gtk_box_pack_start(box, GTK_WIDGET(entry), FALSE, TRUE, extra ? 0 : 6);
+    if(extra)
+        gtk_box_pack_start(box, extra, FALSE, TRUE, 0);
     gtk_widget_show_all(GTK_WIDGET(dlg));
 
     if(has_sel)
@@ -510,7 +551,7 @@ FmPath* fm_select_folder(GtkWindow* parent, const char* title)
     FmPath* path;
     GtkFileChooser* chooser;
     chooser = (GtkFileChooser*)gtk_file_chooser_dialog_new(
-                                        title ? title : _("Please select a folder"),
+                                        title ? title : _("Select Folder"),
                                         parent, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
@@ -621,13 +662,17 @@ static gboolean fm_do_mount(GtkWindow* parent, GObject* obj, MountAction action,
     case EJECT_VOLUME:
         {
             GMount* mnt = g_volume_get_mount(G_VOLUME(obj));
-            prepare_unmount(mnt);
-            g_object_unref(mnt);
+            if (mnt) /* it might be unmounted already */
+            {
+                prepare_unmount(mnt);
+                g_object_unref(mnt);
+            }
             g_volume_eject_with_operation(G_VOLUME(obj), G_MOUNT_UNMOUNT_NONE, op, cancellable, on_mount_action_finished, data);
         }
         break;
     }
 
+    /* FIXME: create progress window with busy cursor */
     if (g_main_loop_is_running(data->loop))
     {
         GDK_THREADS_LEAVE();
@@ -865,11 +910,33 @@ void fm_link_files(GtkWindow* parent, FmPathList* files, FmPath* dest_dir)
  */
 void fm_trash_files(GtkWindow* parent, FmPathList* files)
 {
-    if(!fm_config->confirm_del || fm_yes_no(parent, NULL, _("Do you want to move the selected files to trash can?"), TRUE))
+    FmFileOpsJob *job;
+    char *msg, *name;
+    int len;
+
+    if (fm_config->confirm_trash)
     {
-        FmFileOpsJob* job = fm_file_ops_job_new(FM_FILE_OP_TRASH, files);
-        fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
+        len = fm_path_list_get_length(files);
+        if (len == 1)
+        {
+            name = fm_path_display_basename(fm_path_list_peek_head(files));
+            msg = g_strdup_printf(_("Do you want to move the file '%s' to trash can?"), name);
+            g_free(name);
+        }
+        else
+            msg = g_strdup_printf(dngettext(GETTEXT_PACKAGE,
+                                            "Do you want to move the %d selected file to trash can?",
+                                            "Do you want to move the %d selected files to trash can?",
+                                            (gulong)len), len);
+        if (!fm_yes_no(parent, NULL, msg, TRUE))
+        {
+            g_free(msg);
+            return;
+        }
+        g_free(msg);
     }
+    job = fm_file_ops_job_new(FM_FILE_OP_TRASH, files);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
 }
 
 /**
@@ -909,8 +976,31 @@ static void fm_delete_files_internal(GtkWindow* parent, FmPathList* files)
  */
 void fm_delete_files(GtkWindow* parent, FmPathList* files)
 {
-    if(!fm_config->confirm_del || fm_yes_no(parent, NULL, _("Do you want to delete the selected files?"), TRUE))
-        fm_delete_files_internal(parent, files);
+    char *msg, *name;
+    int len;
+
+    if (fm_config->confirm_del)
+    {
+        len = fm_path_list_get_length(files);
+        if (len == 1)
+        {
+            name = fm_path_display_basename(fm_path_list_peek_head(files));
+            msg = g_strdup_printf(_("Do you want to delete the file '%s'?"), name);
+            g_free(name);
+        }
+        else
+            msg = g_strdup_printf(dngettext(GETTEXT_PACKAGE,
+                                            "Do you want to delete the %d selected file?",
+                                            "Do you want to delete the %d selected files?",
+                                            (gulong)len), len);
+        if (!fm_yes_no(parent, NULL, msg, TRUE))
+        {
+            g_free(msg);
+            return;
+        }
+        g_free(msg);
+    }
+    fm_delete_files_internal(parent, files);
 }
 
 /**
@@ -992,38 +1082,67 @@ void fm_move_or_copy_files_to(GtkWindow* parent, FmPathList* files, gboolean is_
  */
 void fm_rename_file(GtkWindow* parent, FmPath* file)
 {
-    GFile *gf, *parent_gf, *dest;
-    GError* err = NULL;
-    gchar* new_name;
-    GFileMonitor* mon;
+    gchar *old_name, *new_name;
+    FmPathList *files;
+    FmFileOpsJob *job;
 
+    /* NOTE: it's better to use fm_file_info_get_edit_name() to get a name
+       but we cannot get it from FmPath */
+    old_name = fm_path_display_basename(file);
     new_name = fm_get_user_input_rename(parent, _("Rename File"),
                                         _("Please enter a new name:"),
-                                        fm_path_get_basename(file));
+                                        old_name);
+    g_free(old_name);
     if( !new_name )
         return;
-    gf = fm_path_to_gfile(file);
-    parent_gf = g_file_get_parent(gf);
-    dest = g_file_get_child(G_FILE(parent_gf), new_name);
-    if(!g_file_move(gf, dest,
-                G_FILE_COPY_ALL_METADATA|
-                G_FILE_COPY_NO_FALLBACK_FOR_MOVE|
-                G_FILE_COPY_NOFOLLOW_SYMLINKS,
-                NULL, /* make this cancellable later. */
-                NULL, NULL, &err))
-    {
-        fm_show_error(parent, NULL, err->message);
-        g_error_free(err);
-    }
-    else if((mon = fm_monitor_lookup_dummy_monitor(parent_gf)) != NULL)
-    {
-        g_file_monitor_emit_event(mon, gf, NULL, G_FILE_MONITOR_EVENT_DELETED);
-        g_file_monitor_emit_event(mon, dest, NULL, G_FILE_MONITOR_EVENT_CREATED);
-        g_object_unref(mon);
-    }
-    g_object_unref(parent_gf);
-    g_object_unref(dest);
-    g_object_unref(gf);
+    files = fm_path_list_new();
+    fm_path_list_push_tail(files, file);
+    job = fm_file_ops_job_new(FM_FILE_OP_CHANGE_ATTR, files);
+    fm_file_ops_job_set_display_name(job, new_name);
+    g_free(new_name);
+    fm_path_list_unref(files);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
+}
+
+static void _fm_set_file_hidden(GtkWindow *parent, FmPath *file, gboolean hidden)
+{
+    FmPathList *files;
+    FmFileOpsJob *job;
+
+    files = fm_path_list_new();
+    fm_path_list_push_tail(files, file);
+    job = fm_file_ops_job_new(FM_FILE_OP_CHANGE_ATTR, files);
+    fm_file_ops_job_set_hidden(job, hidden);
+    fm_path_list_unref(files);
+    fm_file_ops_job_run_with_progress(parent, job); /* it eats reference! */
+}
+
+/**
+ * fm_hide_file
+ * @parent: a window to place progress dialog over it
+ * @file: the file
+ *
+ * Sets file attribute "hidden" for @file.
+ *
+ * Since: 1.2.0
+ */
+void fm_hide_file(GtkWindow* parent, FmPath* file)
+{
+    _fm_set_file_hidden(parent, file, TRUE);
+}
+
+/**
+ * fm_unhide_file
+ * @parent: a window to place progress dialog over it
+ * @file: the file
+ *
+ * Unsets file attribute "hidden" for @file.
+ *
+ * Since: 1.2.0
+ */
+void fm_unhide_file(GtkWindow* parent, FmPath* file)
+{
+    _fm_set_file_hidden(parent, file, FALSE);
 }
 
 /**
@@ -1068,7 +1187,7 @@ void fm_set_busy_cursor(GtkWidget* widget)
     else
     {
         /* FIXME: how to handle this case? */
-        /* g_debug("not realized"); */
+        g_warning("fm_set_busy_cursor: widget is not realized");
     }
 }
 
@@ -1090,3 +1209,68 @@ void fm_unset_busy_cursor(GtkWidget* widget)
         gdk_window_set_cursor(window, NULL);
     }
 }
+
+static void assign_tooltip_from_action(GtkWidget* widget)
+{
+    GtkAction* action;
+    const gchar * tooltip;
+
+    action = gtk_activatable_get_related_action(GTK_ACTIVATABLE(widget));
+    if (!action)
+        return;
+
+    if (!gtk_activatable_get_use_action_appearance(GTK_ACTIVATABLE(widget)))
+        return;
+
+    tooltip = gtk_action_get_tooltip(action);
+    if (tooltip)
+    {
+        gtk_widget_set_tooltip_text(widget, tooltip);
+        gtk_widget_set_has_tooltip(widget, TRUE);
+    }
+    else
+    {
+        gtk_widget_set_has_tooltip(widget, FALSE);
+    }
+}
+
+static void assign_tooltips_from_actions(GtkWidget* widget)
+{
+    if(G_LIKELY(GTK_IS_MENU_ITEM(widget)))
+    {
+        if(GTK_IS_ACTIVATABLE(widget))
+            assign_tooltip_from_action(widget);
+        widget = gtk_menu_item_get_submenu((GtkMenuItem*)widget);
+        if(widget)
+            assign_tooltips_from_actions(widget);
+    }
+    else if (GTK_IS_CONTAINER(widget))
+    {
+        gtk_container_forall((GtkContainer*)widget,
+                             (GtkCallback)assign_tooltips_from_actions, NULL);
+    }
+}
+
+/**
+ * fm_widget_menu_fix_tooltips
+ * @menu: a #GtkMenu instance
+ *
+ * Fix on GTK bug: it does not assign tooltips of menu items from
+ * appropriate #GtkAction objects. This API assigns them instead.
+ *
+ * Since: 1.2.0
+ */
+void fm_widget_menu_fix_tooltips(GtkMenu *menu)
+{
+    GtkWidget *parent;
+    GtkSettings *settings;
+    gboolean tooltips_enabled;
+
+    g_return_if_fail(GTK_IS_MENU(menu));
+    parent = gtk_menu_get_attach_widget(menu);
+    settings = parent ? gtk_settings_get_for_screen(gtk_widget_get_screen(parent))
+                      : gtk_settings_get_default();
+    g_object_get(G_OBJECT(settings), "gtk-enable-tooltips", &tooltips_enabled, NULL);
+    if(tooltips_enabled)
+        assign_tooltips_from_actions(GTK_WIDGET(menu));
+}
